unit DropDemoAppInterface;

{---------------------------------------------------------------
Purpose: Implements App-to-TDropTarget interface for DropDemo app.

Author:  Graham Wideman
         graham@wideman-one.com
         www.wideman-one.com

TDropDemoAppInterface
-------------
Demos how to subclass TGWDropAppInterface and implement the
app-specific portion of an OLE-drop-accepting application.

Revision History
----------------
98-06-26 GW Original

To-do list
----------
-- Add format details block to Data_ContentToLines
-- Add save-to-file capability
 
---------------------------------------------------------------}

interface
uses
  Windows, ActiveX,
  Classes, forms,
  GWDropTarget;

Const
  FormatCountMax = 20;

Type
  //------------------------------------
  TGWDropRecord = record
  //------------------------------------
    FormatEtc: TFormatEtc;
    StgMedium: TStgMedium;
    datasize : integer;
  end;

  TGWDropRecordArray = array [0..FormatCountMax-1] of TGWDropRecord;

  //------------------------------------
  TDropDemoAppInterface = class(TGWDropAppInterface)
  //------------------------------------
  private
    //--------------------------------------------
    // App-specific data
    //--------------------------------------------
    // fields public due only to laziness...
    FDrop_StringSave: string;

    FFormatCount: integer;
    FDropRecord  : TGWDropRecordArray;

  public
    property FormatCount: integer read FFormatCount;
    property DropRecord:  TGWDropRecordArray read FDropRecord;


    constructor Create(AOwner: TForm); virtual; // Monitor = true means use monitoring

    //---------------------------------------
    // Drop Event functions
    // We will override these since we want
    // fairly major change of function.
    //---------------------------------------

    function DragEnter(const dataObj: IDataObject; grfKeyState: Longint;
      pt: TPoint; var dwEffect: Longint): HResult; override;

    function DragOver(grfKeyState: Longint; pt: TPoint;
      var dwEffect: Longint): HResult; override;

    function DragLeave: HResult; override;

    function Drop(const dataObj: IDataObject; grfKeyState: Longint; pt: TPoint;
      var dwEffect: Longint): HResult; override;

    //----------------------------------
    // Drop Event Helpers
    //----------------------------------
    // Not used or overriden
    // function  Drop_Logic...
    // procedure Drop_Feedback...

    //--------------------------------------------
    // App-specific data and UI-handler functions
    //--------------------------------------------
    procedure Data_Clear;
    procedure Data_HandleDrop(dataObj: IDataObject);
    procedure Data_ContentToLines(FmtNum: integer; L: TStrings; ByteLimit: integer);

  end;

//--------------------------------
implementation
//--------------------------------
uses sysutils, stdctrls, graphics,
     GWDropDemo_Main;  // so we can fiddle with form directly

//--------------------------------------------
constructor TDropDemoAppInterface.Create(AOwner: TForm); // Monitor = true means use monitoring
//--------------------------------------------
Var N: integer;
Begin
  inherited Create(AOwner);

  FFormatCount    := 0;
  For N := 0 to FormatCountMax - 1 do
  Begin
    FDropRecord[N].StgMedium.tymed := TYMED_NULL; // indicate not in use
    FDropRecord[N].datasize := 0;
    // don't care about FormatEtc
  end;
end;

//--------------------------------------------
function TDropDemoAppInterface.DragEnter(
            const dataObj: IDataObject;
              grfKeyState: Longint;
                       pt: TPoint;
              var dwEffect: Longint): HResult;
//--------------------------------------------
Var
  FormatEtc  : TFormatEtc;
  Rslt       : HResult;
  HDropAvail: Boolean;
  S: string;
Begin
  // we're going to accept anything, but just for demo, we'll show
  // how to use QueryGetData
  With FormatEtc do
  Begin
    cfFormat := CF_HDROP;
    ptd      := nil;
    dwAspect := DVASPECT_CONTENT;
    lindex   := -1;
    tymed    := TYMED_HGLOBAL;
  end;

  Rslt := dataObj.QueryGetData(FormatEtc);
  HDropAvail := (Rslt = S_OK);

  // We could use this result to accept or reject the drop,
  // but instead we'll just use it for a bit of feedback

  If HDropAvail then
    S := 'HDrop Available' else
    S := 'No HDrop';

  //-------------------------------------
  // Feedback
  //-------------------------------------
  FDrop_ColorSave  := (DropTarget.DropWinControl as TStaticText).Color;
  (DropTarget.DropWinControl as TStaticText).Color := clLime;

  FDrop_StringSave := (DropTarget.DropWinControl as TStaticText).Caption;
  (DropTarget.DropWinControl as TStaticText).Caption := S;

  FDrop_FeedbackOn := true;

  dwEffect := Drop_StdEffect(grfKeyState);
  Result := S_OK;
end;

//--------------------------------------------
function TDropDemoAppInterface.DragOver(
              grfKeyState: Longint;
                       pt: TPoint;
             var dwEffect: Longint): HResult;
//--------------------------------------------
Begin
  // No particular data processing
  dwEffect := Drop_StdEffect(grfKeyState);

  // No particular feedback

  Result := S_OK;
end;

//--------------------------------------------
function TDropDemoAppInterface.DragLeave: HResult;
//--------------------------------------------
Begin

  //-------------------------------------
  // Feedback
  //-------------------------------------
  if FDrop_FeedbackOn then
  Begin
    (DropTarget.DropWinControl as TStaticText).Color   := FDrop_ColorSave;
    (DropTarget.DropWinControl as TStaticText).Caption := FDrop_StringSave;
  end;
  FDrop_FeedbackOn := false;

  Result := S_OK;
end;

//--------------------------------------------
function TDropDemoAppInterface.Drop(
            const dataObj: IDataObject;
              grfKeyState: Longint;
                       pt: TPoint;
             var dwEffect: Longint): HResult;
//--------------------------------------------
Begin
  //-------------------------------------
  // Deal with the data
  //-------------------------------------
  Data_HandleDrop(dataObj);

  //-------------------------------------
  // Feedback
  //-------------------------------------
  if FDrop_FeedbackOn then
  Begin
    (DropTarget.DropWinControl as TStaticText).Color   := FDrop_ColorSave;
    (DropTarget.DropWinControl as TStaticText).Caption := FDrop_StringSave;
  end;
  FDrop_FeedbackOn := false;


  dwEffect := DROPEFFECT_COPY; // so we tell source app not to delete
  Result := S_OK;
end;

//--------------------------------------------
procedure TDropDemoAppInterface.Data_Clear;
//--------------------------------------------
Var N: integer;
Begin
  FFormatCount    := 0;
  For N := 0 to FormatCountMax - 1 do
  Begin
    If (FDropRecord[N].StgMedium.tymed <> TYMED_NULL) and
       (FDropRecord[N].StgMedium.unkForRelease = nil) then { we're responsible for freeing }
       ReleaseStgMedium(FDropRecord[N].StgMedium);

    FDropRecord[N].StgMedium.tymed := TYMED_NULL; // indicate not in use
  end;
end;

//-------------------------------------------------------
procedure TDropDemoAppInterface.Data_HandleDrop(dataObj: IDataObject);
//-------------------------------------------------------
Var
  Rslt: HResult;
  EnumFormatEtc: IEnumFormatEtc;
  FetchedCount: longint;
  OK: boolean;
  FE: TFormatEtc;
  N: integer;
  Handle: HGlobal;
Const FetchRequest = 1;
Begin
  Data_Clear;

  //---------------------------------------
  // Get all the EnumFormats
  //---------------------------------------
  Rslt := dataObj.EnumFormatEtc(DATADIR_GET, EnumFormatEtc);
  FFormatCount := 0;
  Rslt := EnumFormatEtc.Reset;
  Rslt := S_OK;
  While (Rslt = S_OK) and (FFormatCount < FormatCountMax-1) do
  Begin
    Rslt := EnumFormatEtc.Next(FetchRequest, FE, @FetchedCount);
    If Rslt = S_OK then
    Begin
      FDropRecord[FFormatCount].FormatEtc.cfFormat := FE.cfFormat;
      FDropRecord[FFormatCount].FormatEtc.ptd      := FE.ptd;
      FDropRecord[FFormatCount].FormatEtc.dwAspect := FE.dwAspect;
      FDropRecord[FFormatCount].FormatEtc.lindex   := FE.lindex;
      FDropRecord[FFormatCount].FormatEtc.tymed    := FE.tymed;
      Inc(FFormatCount);
    end;
  end;

  //-------------------------------------------------------------
  // Get all HGLOBAL data (StgMediums with global-allocated data)
  //-------------------------------------------------------------
  For N := 0 to FFormatCount - 1 do
  Begin
    If (DropRecord[N].FormatEtc.tymed and TYMED_HGLOBAL) <> 0 then
    Begin
      FE.cfFormat  :=  FDropRecord[N].FormatEtc.cfFormat;
      FE.ptd       :=  FDropRecord[N].FormatEtc.ptd     ;
      FE.dwAspect  :=  FDropRecord[N].FormatEtc.dwAspect;
      FE.lindex    :=  FDropRecord[N].FormatEtc.lindex;
      FE.tymed     :=  TYMED_HGLOBAL;
      Rslt := dataObj.GetData(FE, FDropRecord[N].StgMedium);
      If Rslt = S_OK then
      Begin
        Handle := FDropRecord[N].StgMedium.hGlobal;
        FDropRecord[N].datasize := GlobalSize(Handle);
      end;
    end;
  end;

  (FOwnerForm as Tfm_Main).OnOLEDrop;
end;

//-------------------------------
procedure TDropDemoAppInterface.Data_ContentToLines(
          FmtNum: integer; L: TStrings; ByteLimit: integer);
//-------------------------------
Var
  FmtName: String;
  S, S1, S2, S3: string;
  C: char;
  CharNum, datasize, CharCountThisLine: integer;
  ptr: PChar;
Const
  CharPerLine = 16;
  MaxDataSize = 4096;

  procedure AddLine;
  Begin
    While Length(S2) < 48 do S2 := S2 + ' ';
    S := S1 + ' | ' + S2 + ' | ' + S3;
    L.Add(S);
    S1 := IntToHex(CharNum+1,6);  S2 := '';  S3 := '';
    CharCountThisLine := 0;
  end;

Label the_end;
Begin
  If (FmtNum < 0) or (FmtNum > FormatCountMax-1) then
  Begin
    S := 'Error: Attempt to add format: ' + IntToStr(FmtNum);
    L.Add(S);
    goto the_end;
  end;

  FmtName := cfFormatCaption(DropRecord[FmtNum].FormatEtc.cfFormat);
  L.Add('Format: '+ FmtName);

  If DropRecord[FmtNum].StgMedium.tymed <> TYMED_HGLOBAL then
  Begin
    S := 'Format does not use a global memory handle.';
    L.Add(S);
    goto the_end;
  end;

  datasize := DropRecord[FmtNum].datasize;
  ptr      := GlobalLock(DropRecord[FmtNum].StgMedium.HGlobal);
  try
    CharNum           := 0;
    CharCountThisLine := 0;
    S1 := IntToHex(CharNum,6);  S2 := '';  S3 := '';
    While (CharNum < datasize) and (CharNum < ByteLimit) do
    Begin
      Inc(CharCountThisLine);
      C := ptr[CharNum];
      S2 := S2 + ' ' + IntToHex(Ord(C),2);
      If (C >= ' ') and (Ord(C) <= $7F) then S3 := S3 + C else S3 := S3 + '.';

      If CharCountThisLine = CharPerLine then AddLine;
      Inc(CharNum);
    end;
    If CharCountThisLine > 0 then AddLine;
    If CharNum = ByteLimit then L.Add('...more...');
  finally
    GlobalUnLock(DropRecord[FmtNum].StgMedium.HGlobal);
  end;
the_end:
end;

end.

