unit GWDropTarget;
{---------------------------------------------------------------
Purpose: Implements essential foundation for OLE Drop Target.

Author:  Graham Wideman
         graham@wideman-one.com
         www.wideman-one.com

TGWDropTarget
-------------
Implements:
1)  generic IDropTarget interface.
2)  generic monitoring for testing
Should be useable as-is in apps... place all app-specific functionality in
descendant of TGWDropAppInterface

TGWDropAppInterface
--------------------
Implements:
1) Lifetime management of a TGWDropTarget
2) monitoring functions
3) default dumb implementations for IDropTarget events
Override this class to provide app-specific functions

Revision History
----------------
98-06-26 GW factored functionality better
98-06-22 GW revised LifeMgr and Monitor architecture
98-06-20 GW original

To-do list
----------

---------------------------------------------------------------}

//-----------------------
interface
//-----------------------
uses Windows, Messages, SysUtils, Classes, graphics,
     ActiveX, stdctrls, controls, forms, grids;

Type
  //---------------------------------------
  // forward declarations
  //---------------------------------------
  TGWDropTarget    = class;

  // Life states
  Tgwdt_ls = (gwdt_ls_start, gwdt_ls_exists, gwdt_ls_locked, gwdt_ls_regd);
  // Drop phases
  Tddp = (ddp_DragEnter, ddp_DragOver, ddp_DragLeave, ddp_Drop);

  //---------------------------------------
  TGWDropAppInterface = class
  {---------------------------------------
  Implements:
    TGWDropTarget lifetime and pointer.

  Default Implementations:
    Drop hit detect
    Drop accept/reject
    UI feedback
    Drop data processing
  ----------------------------------------}
  protected
    FOwnerForm   : TForm;
    FDropTarget  : TGWDropTarget;      // The object being managed
                                       // DropTarget will nil this when destroyed
    FMon_Enable  : boolean;

    //--------------------------------
    // Default drop helper fields
    //--------------------------------

    FDrop_ColorSave  : TColor;
    FDrop_FeedbackOn : Boolean;

  public
    property  DropTarget: TGWDropTarget read FDropTarget;

    //---------------------------------------
    // Production functions
    //---------------------------------------
    // Create in  OwnerForm.OnCreate, Destroy in OwnerForm.OnDestroy
    constructor Create(AOwner: TForm); virtual;
    destructor  Destroy; override;

    //---------------------------------------
    // Drop Target Lifetime management
    // Should be no need to override
    //---------------------------------------
    function   DropTarget_Create(
                    AWinControl: TWinControl;
                    Reg: Boolean): HResult;
    function   DropTarget_Exists: Boolean;
  protected
    procedure  DropTarget_Forget;  // used by DropTarget.Destroy;
  public
    function   DropTarget_LifeState: Tgwdt_ls;

    //---------------------------------------
    // Monitoring functions (generic)
    //---------------------------------------
    property  Mon_Enable: boolean Read FMon_Enable Write FMon_Enable;
    function  Mon_GridInit(sg: TStringGrid; StartRow: integer): integer; virtual;
    function  Mon_GridUpdate(sg: TStringGrid; StartRow: integer): integer; virtual;
    procedure Mon_ClearCallCounts; virtual;

    //------------------------------------------------
    // Drop Event functions (generic)
    // ie: TGWDropTarget delegates to here
    // For major app-specific changes in behavior, override these
    // for minor ones override the Helpers instead.
    //------------------------------------------------
    function DragEnter(const dataObj: IDataObject; grfKeyState: Longint;
      pt: TPoint; var dwEffect: Longint): HResult; virtual;

    function DragOver(grfKeyState: Longint; pt: TPoint;
      var dwEffect: Longint): HResult; virtual;

    function DragLeave: HResult; virtual;

    function Drop(const dataObj: IDataObject; grfKeyState: Longint; pt: TPoint;
      var dwEffect: Longint): HResult; virtual;

    //---------------------------------------
    // Drop Event Helpers
    // May override these for app-specific behavior, if
    // still using above delegation functions
    //---------------------------------------
    function  Drop_StdEffect(grfKeyState: Longint): Longint; virtual;
    function  Drop_Logic(phase: Tddp; dataObj: IDataObject; grfKeyState: Longint;
                          pt: TPoint): Longint; virtual; //dwEffect
    procedure Drop_Feedback(phase: Tddp; pt: TPoint; dwEffect: longint); virtual;

  end;

  //------------------------------------------------
  // Monitor Counters
  Tdmc = (dmc_DragEnter, dmc_DragOver, dmc_DragLeave, dmc_Drop);

  //------------------------------------
  TGWDropTarget = class (TInterfacedObject, IDropTarget)
  {------------------------------------
  Implements:
    -- Generic drop functions, delegates app-specific
       stuff to descendant of TGWDropAppInterface.
    -- Monitor functions
  -------------------------------------}
  private
    //-----------------------------
    // DropTarget features
    //-----------------------------
    FDropHWND       : HWND;        // must remember in order to Revoke
    FDropWinControl : TWinControl;

    //-----------------------------
    // State management
    //-----------------------------
    FDropAppInterface  : TGWDropAppInterface;
    FLifeState         : Tgwdt_ls;
    FEnable            : boolean;  // for temporary enable/disable

    //---------------------------------
    //  Monitoring
    //---------------------------------
    Mon_Counts: array[dmc_DragEnter..dmc_Drop] of integer;

    //---------------------------------
    //  Property procs
    //---------------------------------
    procedure SetLifeState(Value: Tgwdt_ls);
    procedure SetEnable(Value: Boolean);

  public
    //-----------------------------
    // DropTarget functions
    //-----------------------------
    property DropHWND       : HWND         read FDropHWnd;
    property DropWinControl : TWinControl  read FDropWinControl;

    property LifeState: Tgwdt_ls Read FLifeState Write SetLifeState;
    property Enable   : Boolean Read FEnable     Write SetEnable;

    constructor Create(AWinControl: TWinControl; ADropAppInterface: TGWDropAppInterface); virtual;
    destructor  Destroy; override;

    function ToState_Exists : HResult;
    function ToState_Locked : HResult;
    function ToState_Regd   : HResult;

    //---------------------------------
    //  Monitor and Test functions
    //---------------------------------
    class function  Mon_GridInit(sg: TStringGrid; StartRow: integer): integer;
    function  Mon_GridUpdate(sg: TStringGrid; StartRow: integer): integer;
    procedure Mon_ClearCallCounts;
  protected
    function  Mon_Enable: boolean;
    procedure Mon_IncCount(Which: Tdmc);

  public
    function  TestAddRef : integer;
    function  TestRelease: integer;

    //---------------------------------
    //   IDropTarget
    //---------------------------------
    function DragEnter(const dataObj: IDataObject; grfKeyState: Longint;
      pt: TPoint; var dwEffect: Longint): HResult; stdcall;

    function DragOver(grfKeyState: Longint; pt: TPoint;
      var dwEffect: Longint): HResult; stdcall;

    function DragLeave: HResult; stdcall;

    function Drop(const dataObj: IDataObject; grfKeyState: Longint; pt: TPoint;
      var dwEffect: Longint): HResult; stdcall;
  end;

  //---------------------------------------------
  // Hack to get at color property...
  // See Cantu etc "Delphi Developer's Handbook"
  //---------------------------------------------
  THackControl = class(TControl)
  end;

//-------------------------------------
// Misc Utility
//-------------------------------------
  function cfFormatCaption(I: integer): string;
  function dwAspectCaption(I: integer): string;
  function tymedCaption(I: integer): string;

//-----------------------
implementation
//-----------------------

//-------------------------------------------------------
//                TGWDropAppInterface
//-------------------------------------------------------

//-----------------------------------------
constructor TGWDropAppInterface.Create(AOwner: TForm);
//-----------------------------------------
Begin
  inherited Create;
  FOwnerForm         := AOwner;
  FDropTarget        := nil;
  FMon_Enable        := false;
  FDrop_FeedbackOn   := false;
end;

//-----------------------------------------
destructor TGWDropAppInterface.Destroy;
//-----------------------------------------
Begin
  If Assigned(FDropTarget) then FDropTarget.Destroy;  // will nil FDropTarget
  inherited Destroy;
end;

//------------------------------------------------------
function  TGWDropAppInterface.DropTarget_Create(
                    AWinControl: TWinControl;
                    Reg: Boolean): HResult;
//------------------------------------------------------
Begin
  Result := E_UNEXPECTED;
  FDropTarget := TGWDropTarget.Create(AWinControl, Self);
  If Assigned(FDropTarget) then
  Begin
    Result := S_OK;
    If Reg then { go ahead and lock and register }
    Begin
      Result := DropTarget.ToState_Regd;
    end;
  end;
end;

//------------------------------------------------------
function TGWDropAppInterface.DropTarget_Exists: Boolean;
//------------------------------------------------------
Begin
  Result := Assigned(FDropTarget);
end;

//------------------------------------------------------
procedure TGWDropAppInterface.DropTarget_Forget;
//------------------------------------------------------
Begin
  FDropTarget := nil;
end;
//-----------------------------------------
function  TGWDropAppInterface.DropTarget_LifeState: Tgwdt_ls;
//-----------------------------------------
Begin
  If DropTarget_Exists then
  Begin
    Result := DropTarget.LifeState;
  end else
  Begin
    Result := gwdt_ls_Start;
  end;
end;

//---------------------------------------------------------------
//     Monitor and Test Functions
//---------------------------------------------------------------

Const
  dai_Rows = 1;
  dai_RowCaptions: array[0..dai_Rows-1] of string = ('Exists');

//----------------------------------------------------
function TGWDropAppInterface.Mon_GridInit(sg: TStringGrid; StartRow: integer): integer;
//----------------------------------------------------
// Returns number of next available row
Var
  N: integer;
  RowNum: integer;
Begin
  if sg.RowCount < StartRow + dai_Rows  then
    sg.RowCount := StartRow + + dai_Rows;

  RowNum := StartRow;

  For N := 0 to dai_Rows-1 do
  Begin
    sg.Cells[0,RowNum] := dai_RowCaptions[N];
    Inc(RowNum)
  end;

  RowNum := TGWDropTarget.Mon_GridInit(sg, RowNum);
  Result := RowNum;
end;

//------------------------------------------------------
function TGWDropAppInterface.Mon_GridUpdate(sg: TStringGrid; StartRow: integer): integer;
//------------------------------------------------------
// Returns number of next available row
Var
  I: integer;
  S: string;
  RowNum: integer;
Begin
  if sg.RowCount < StartRow + dai_Rows  then
    sg.RowCount := StartRow + dai_Rows ;

  RowNum := StartRow;

  If DropTarget_Exists then
  Begin
    sg.Cells[1,RowNum] := '1';
    sg.Cells[2,RowNum] := 'Yes';
    Inc(RowNum);
    RowNum := DropTarget.Mon_GridUpdate(sg, RowNum);
  end else
  Begin
    sg.Cells[1,RowNum] := '0';
    sg.Cells[2,RowNum] := 'No';
    Inc(RowNum);
    While RowNum <= sg.RowCount-1 do
    Begin
      sg.Cells[1,RowNum] := '---';
      sg.Cells[2,RowNum] := '---';
      Inc(RowNum);
    end;
  end;
  Result := RowNum;
end;

//-----------------------------------------
procedure TGWDropAppInterface.Mon_ClearCallCounts;
//-----------------------------------------
Begin
  If DropTarget_Exists then
  Begin
    DropTarget.Mon_ClearCallCounts;
  end;
end;

//------------------------------------------------------
//  Generic handlers for GWDropTarget delegation
//  An actual app will want to override some or all of these.
//  The default functionality is just minimal to make something
//  interesting appear to happen.
//------------------------------------------------------

//------------------------------------------------------
function TGWDropAppInterface.DragEnter(
              const dataObj      : IDataObject;
                    grfKeyState  : Longint;
                    pt           : TPoint;
                var dwEffect     : Longint): HResult;
//------------------------------------------------------
Begin
  // dumb default behavior is to accept everything without
  // checking data type, and give standard feedback according to
  // key shift status
  dwEffect := Drop_Logic(ddp_DragEnter, dataObj, grfKeyState, pt);
  If (dwEffect <> DROPEFFECT_NONE) then
    Drop_Feedback(ddp_DragEnter, pt, dwEffect);
  Result := S_OK;
end;

//------------------------------------------------------
function TGWDropAppInterface.DragOver(
                    grfKeyState : Longint;
                    pt          : TPoint;
                var dwEffect    : Longint): HResult;
//------------------------------------------------------
Begin
  dwEffect := Drop_Logic(ddp_DragOver, nil, grfKeyState, pt);
  If dwEffect <> DROPEFFECT_NONE then Drop_Feedback(ddp_DragOver,pt, dwEffect);
  Result := S_OK;
end;

//------------------------------------------------------
function TGWDropAppInterface.DragLeave: HResult;
//------------------------------------------------------
Var
  pt: TPoint;
Begin
  Drop_Feedback(ddp_DragLeave, pt, DROPEFFECT_NONE);
  Result := S_OK;
end;

//------------------------------------------------------
function TGWDropAppInterface.Drop(
             const dataObj      : IDataObject;
                   grfKeyState  : Longint;
                   pt           : TPoint;
               var dwEffect     : Longint): HResult;
//------------------------------------------------------
Begin
  {------------------------------------------------------
  Override to implement drop processing.
  Normally this would depend at least on gfKeyState and dataObj
  ---------------------------------}

  dwEffect := Drop_Logic(ddp_Drop, dataObj, grfKeyState, pt);
  Drop_Feedback(ddp_Drop, pt, dwEffect);

  {-------------------------------------------------------
  Tell source app this was a COPY, since for test app
  we don't want the source app to delete the data (as would
  happen if the user requested a MOVE.)
  --------------------------------------------------------}

  Result := S_OK;
end;

//---------------------------------------
// Drop Event Helpers
//---------------------------------------

//-------------------------------------------------------
function TGWDropAppInterface.Drop_StdEffect(grfKeyState: Longint): Longint;
//-------------------------------------------------------
Begin
  Result := DROPEFFECT_COPY; // our default. Note, MOVE is Win standard for no keys
  if (grfKeyState and MK_CONTROL)            <> 0 then Result := DROPEFFECT_COPY;
  if (grfKeyState and MK_SHIFT)              <> 0 then Result := DROPEFFECT_MOVE;
  if (grfKeyState and (MK_SHIFT+MK_CONTROL)) <> 0 then Result := DROPEFFECT_LINK;
end;

//-------------------------------------------------------
function TGWDropAppInterface.Drop_Logic(
              phase: Tddp;
            dataObj: IDataObject;
        grfKeyState: Longint;
                 pt: TPoint): Longint;
//-------------------------------------------------------
Begin
  if phase <> ddp_Drop then
    Result := Drop_StdEffect(grfKeyState) else
    Result := DROPEFFECT_COPY; // so default is to tell source app not to delete anything
end;

//------------------------------------------------------------------------
procedure TGWDropAppInterface.Drop_Feedback(phase: Tddp;
                          pt: TPoint; dwEffect: longint);
//------------------------------------------------------------------------

Begin
  Case phase of
    ddp_DragEnter: Begin
      if dwEffect <> DROPEFFECT_NONE then
      Begin
        FDrop_ColorSave := THackControl(DropTarget.FDropWinControl).Color;
        THackControl(DropTarget.FDropWinControl).Color := clLime;
        FDrop_FeedbackOn := true;
      end;
    end;
    ddp_DragOver : { nothing } ;
    ddp_DragLeave: Begin
      if FDrop_FeedbackOn then
        THackControl(DropTarget.FDropWinControl).Color := FDrop_ColorSave;
      FDrop_FeedbackOn := false;
    end;
    ddp_Drop     : Begin
      if FDrop_FeedbackOn then
        THackControl(DropTarget.FDropWinControl).Color := FDrop_ColorSave;
      FDrop_FeedbackOn := false;
    end;
  end; { case }
end;

//------------------------------------------------------
//     TGWDropTarget.
//------------------------------------------------------

//------------------------------------------------------
constructor TGWDropTarget.Create(AWinControl: TWinControl; ADropAppInterface: TGWDropAppInterface);
//------------------------------------------------------
Begin
  inherited Create;
  FDropHWND          := AWinControl.Handle;
  FDropWinControl    := AWinControl;
  FDropAppInterface  := ADropAppInterface;
  FLifeState         := gwdt_ls_exists;
  FEnable            := true;
  Mon_ClearCallCounts;
end;

//------------------------------------------------------
destructor TGWDropTarget.Destroy;
//------------------------------------------------------
{ This version of Destroy is safe to call from:
 a) an application (eg: during shutdown)
 b) the system, where the application has explicitly
    stepped through ToState functions to Revoke/unlock
 c) the system if Release is used on the object (improperly)
    and Revoked, unlock have not been performed

 Complexities:
 1. We must Revoke and unlock, but we must ensure that
    these functions do not call Destroy (ie: make sure RefCount
    doesn't go to zero.) because...
 2. Can't allow Destroy to reenter, because no way to prevent
    "inner" Destroy from freeing the pointer, resulting
    in exception when outer Destroy tries to do likewise.
-------------------------------------------------------}
Var
  Rslt: HResult;
Begin
  // Report My Destroy;
  If Assigned(FDropAppInterface) then
  Begin
    FDropAppInterface.DropTarget_Forget;
  end;
  // make sure to Revoke
  if FLifeState > gwdt_ls_Locked then
  Begin
    While RefCount < 2 do _AddRef;   // avoid retriggering destroy
    ActiveX.RevokeDragDrop(FDropHWND);
    FLifeState := gwdt_ls_Locked;
  end;
  // make sure to unlock
  if FLifeState > gwdt_ls_Exists then
  Begin
    While RefCount < 2 do _AddRef;   // avoid retriggering destroy
    Rslt := ActiveX.CoLockObjectExternal(
      Self,    // unlock me
      false,   // false = unlock
      false    // fLastUnlockReleases false = only release once
    );
    FLifeState := gwdt_ls_Exists;
  end;
  // at this point RefCount may not be zero, but we don't care since we're
  // about to delete the object.
  inherited Destroy;
end;

//------------------------------------------------------
function TGWDropTarget.ToState_Exists : HResult;
//------------------------------------------------------
Begin
  Result := S_OK;
  If LifeState = gwdt_ls_Regd then Result := ToState_Locked;

  If LifeState = gwdt_ls_Locked then
  Begin
    LifeState := gwdt_ls_Exists;   // up front so Destroy won't unlock again.
    Result := ActiveX.CoLockObjectExternal(
      Self,    // unlock me
      false,   // false = unlock
      true     // fLastUnlockReleases
    );
  end;
end;
//------------------------------------------------------
function TGWDropTarget.ToState_Locked : HResult;
//------------------------------------------------------
Begin
  Result := S_OK;

  If LifeState = gwdt_ls_Exists then
  Begin
    Result := ActiveX.CoLockObjectExternal(
      Self,     // unlock me
      true,     // false = unlock
      false);   // fLastUnlockReleases	N/A on lock
    If Result = S_OK then LifeState := gwdt_ls_Locked;
  end;

  If LifeState = gwdt_ls_Regd then
  Begin
    While RefCount < 2 do _AddRef; // avoid premature destroy
    Result := ActiveX.RevokeDragDrop(FDropHWND);
    If Result = S_OK then LifeState := gwdt_ls_Locked;
  end;
end;

//------------------------------------------------------
function TGWDropTarget.ToState_Regd   : HResult;
//------------------------------------------------------
Begin
  Result := S_OK;
  If LifeState = gwdt_ls_Exists then Result := ToState_Locked;
  If LifeState = gwdt_ls_Locked then
  Begin
    Result := ActiveX.RegisterDragDrop(FDropHWND, Self);
    If Result = S_OK then LifeState := gwdt_ls_Regd;
  end;
end;

//--------------------------------------------
procedure TGWDropTarget.SetLifeState(Value: Tgwdt_ls);
//--------------------------------------------
Begin
  FLifeState := Value;
end;
//--------------------------------------------
procedure TGWDropTarget.SetEnable(Value: Boolean);
//--------------------------------------------
Begin
  FEnable := Value;
end;

//--------------------------------------------
//  Monitor and Test
//--------------------------------------------

//----------------------------------------------------
function TGWDropTarget.Mon_Enable: boolean;
//----------------------------------------------------
Begin
  Result := FDropAppInterface.Mon_Enable;
end;

//----------------------------------------------------
Const
  dt_Rows = 7;
  dt_RowCaptions: array[0..dt_Rows-1] of string = (
    'LifeState', 'RefCount' ,'Enabled'   ,
    'DragEnter', 'DragOver' , 'DragLeave', 'Drop');

//----------------------------------------------------
class function TGWDropTarget.Mon_GridInit(sg: TStringGrid; StartRow: integer): integer;
//----------------------------------------------------
// Returns number of next available row
Var
  N: integer;
  RowNum: integer;
Begin
  if sg.RowCount < StartRow + dt_Rows  then
    sg.RowCount := StartRow + dt_Rows;

  RowNum := StartRow;

  For N := 0 to dt_Rows-1 do
  Begin
    sg.Cells[0,RowNum] := dt_RowCaptions[N];
    Inc(RowNum)
  end;
  Result := RowNum;
end;

//------------------------------------------------------
function TGWDropTarget.Mon_GridUpdate(sg: TStringGrid; StartRow: integer): integer;
//------------------------------------------------------
// Returns number of next available row
Var
  I: integer;
  S, S2: string;
  RowNum: integer;
  mc: Tdmc;
Begin
  if sg.RowCount < StartRow + dt_Rows then
    sg.RowCount := StartRow + dt_Rows;

  RowNum := StartRow;

  // 'LifeState', 'RefCount' , 'Enabled'
  Case LifeState of
    gwdt_ls_start : S := 'Start';
    gwdt_ls_exists: S := 'Exists';
    gwdt_ls_locked: S := 'Locked';
    gwdt_ls_regd  : S := 'Regd';
    else S :=  'Error';
  end; { case }
  sg.Cells[1, RowNum] := IntToStr(Ord(LifeState));
  sg.Cells[2, RowNum] := S;
  Inc(RowNum);

  sg.Cells[1, RowNum] := IntToStr(RefCount);
  sg.Cells[2, RowNum] := '';
  Inc(RowNum);

  If Enable then
  Begin
     S  := 'True';
     S2 := 'Yes';
  end else
  Begin
     S  := 'False';
     S2 := 'No';
  end;
  sg.Cells[1, RowNum] := S;
  sg.Cells[2, RowNum] := '';
  Inc(RowNum);

  // 'DragEnter', 'DragOver' , 'DragLeave', 'Drop'
  For mc := dmc_DragEnter to dmc_Drop do
  Begin
    sg.Cells[1, RowNum] := IntToStr(Mon_Counts[mc]);
    sg.Cells[2, RowNum] := '';
    Inc(RowNum);
  end;

  Result := RowNum;
end;

//----------------------------------------------------
procedure TGWDropTarget.Mon_IncCount(Which: Tdmc);
//----------------------------------------------------
Begin
  If Mon_Enable then Inc(Mon_Counts[Which]);
end;

//--------------------------------------------
procedure TGWDropTarget.Mon_ClearCallCounts;
//--------------------------------------------
Var
  N: Tdmc;
Begin
  For N := dmc_DragEnter to dmc_Drop do Mon_Counts[N] := 0;
end;

//--------------------------------------------
function  TGWDropTarget.TestAddRef : integer;
//--------------------------------------------
Begin
  Result := _AddRef;
end;
//--------------------------------------------
function  TGWDropTarget.TestRelease: integer;
//--------------------------------------------
Begin
  Result := _Release;
end;

//----------------------------------------------
// from IDropTarget
// Delegate to TGWDropAppInterface, or decendant
//----------------------------------------------
//------------------------------------------------------
function TGWDropTarget.DragEnter(
              const dataObj      : IDataObject;
                    grfKeyState  : Longint;
                    pt           : TPoint;
                var dwEffect     : Longint): HResult;
//------------------------------------------------------
Begin
  Mon_IncCount(dmc_DragEnter);
  Result := FDropAppInterface.DragEnter(dataObj, grfKeyState, pt, dwEffect);
end;

//------------------------------------------------------
function TGWDropTarget.DragOver(
                    grfKeyState : Longint;
                    pt          : TPoint;
                var dwEffect    : Longint): HResult;
//------------------------------------------------------
Begin
  Mon_IncCount(dmc_DragOver);
  Result := FDropAppInterface.DragOver(grfKeyState, pt, dwEffect);
end;

//------------------------------------------------------
function TGWDropTarget.DragLeave: HResult;
//------------------------------------------------------
Begin
  Mon_IncCount(dmc_DragLeave);
  Result := FDropAppInterface.DragLeave;
end;

//------------------------------------------------------
function TGWDropTarget.Drop(
             const dataObj      : IDataObject;
                   grfKeyState  : Longint;
                   pt           : TPoint;
               var dwEffect     : Longint): HResult;
//------------------------------------------------------
Begin
  Mon_IncCount(dmc_Drop);
  Result := FDropAppInterface.Drop(dataObj, grfKeyState, pt, dwEffect);
end;

//---------------------------------------------------------------
//                 Misc Utility
//---------------------------------------------------------------
//-------------------------------------
function cfFormatCaption(I: integer): string;
//-------------------------------------
Const
  StdMin = 1; StdMax = 17;
  StdFormats: array[StdMin..StdMax] of string = (
    'CF_TEXT',   'CF_BITMAP' , 'CF_METAFILEPICT', 'CF_SYLK',
    'CF_DIF' ,   'CF_TIFF'   , 'CF_OEMTEXT'     , 'CF_DIB',
    'CF_PALETTE','CF_PENDATA', 'CF_RIFF'        , 'CF_WAVE',
    'CF_UNICODETEXT', 'CF_ENHMETAFILE', 'CF_HDROP', 'CF_LOCALE',
    'CF_MAX');
Var
  TempStr: array[0..50] of char;
  pS: PChar;
  Len: integer;
Begin
  Result := '';
  pS := @TempStr;
  Len := GetClipboardFormatName(I, pS, 50);
  If Len > 0 then Result := pS;

  If Result = '' then
  Begin
    Case I of
      StdMin..StdMax: Result := StdFormats[I];
    end; { case }
  end;
  If Result = '' then
  Begin
    Result := '??? ' + IntToStr(I); // default
  end;
end;

//-------------------------------------
function dwAspectCaption(I: integer): string;
//-------------------------------------
Begin
  Result := '??? ' + IntToStr(I); // default
  Case I of
    DVASPECT_CONTENT   : Result := 'CONTENT';
    DVASPECT_THUMBNAIL : Result := 'THUMBNAIL';
    DVASPECT_ICON      : Result := 'ICON';
    DVASPECT_DOCPRINT  : Result := 'DOCPRINT';
  end; { case }
end;

//-------------------------------------
function tymedCaption(I: integer): string;
//-------------------------------------
{ Note, these look suspiciously like they might sometimes be
OR'ed together, in which case this simple-minded case statement
won't return the desired names... }
Begin
  Result := '??? ' + IntToStr(I); // default
  Case I of
    TYMED_NULL     : Result := 'NULL';
    TYMED_HGLOBAL  : Result := 'HGLOBAL';
    TYMED_FILE     : Result := 'FILE';
    TYMED_ISTREAM  : Result := 'ISTREAM';
    TYMED_ISTORAGE : Result := 'ISTORAGE';
    TYMED_GDI      : Result := 'GDI';
    TYMED_MFPICT   : Result := 'MFPICT';
    TYMED_ENHMF    : Result := 'ENHMF';
  end; { case }
end;


//-----------------------
initialization
//-----------------------
  OleInitialize (Nil);

//-----------------------
finalization
//-----------------------
  OleUninitialize;
end.

